# pragma implementation

# include	"../../header.h"
# include	"df.h"

	df::df(const char* dname,const char* fname)
{
	char	name[MaxString];

	sprintf(name,"%s/curr/%s",dname,fname);
	access = 0;
	page_count = 0;
	for (int i = 0; i < DFPoolSize; ++i) {
		pool[i].access = -1;
		pool[i].pno = -1;
		pool[i].modified = False;
	}
	file.open(name,(ios::in)|(ios::out)|(ios::nocreate));
	if (file.rdstate() == ios::goodbit)
		file.read((char*) &page_count, sizeof(page_count));
	else	file.open(name,(ios::in)|(ios::out));
	check(file.rdstate() == ios::goodbit,"df::df");
}

	df::~df()
{
	file.seekp(0,ios::beg);
	file.write((char*) &page_count, sizeof(page_count));
	for (int i = 0; i < DFPoolSize; ++i)
		if (pool[i].modified)
			write(pool[i].pno,pool[i].page);
	file.close();
	check(file.rdstate() == ios::goodbit,"df::~df");
};

void	df::read(int pno,DFPage& p)
{
	DiskPage	d;
	register int	i;

	file.seekg(sizeof(page_count) + (pno*sizeof(DiskPage)),ios::beg);
	file.read((char*) &d, sizeof(DiskPage));
	check(file.rdstate() != (ios::eofbit),"df::read (ios::eofbit)");
	check(file.rdstate() != (ios::badbit),"df::read bad");
	check(file.rdstate() != (ios::failbit),"df::read fail");
	check(file.rdstate() == ios::goodbit,"df:read");
	p.free = d.free;
	p.info = d.info;
	for (i = 0; i < d.free; ++i)
		p.data[i].set(d.tag[i],d.value[i]);
};

void	df::write(int pno,DFPage& p)
{
	DiskPage	d;
	register int	i;

	d.free = p.free;
	d.info = p.info;
	for (i = 0; i < p.free; ++i) {
		d.tag[i] =   p.data[i].tag();
		d.value[i] = p.data[i].value();
	}
	file.seekp(sizeof(page_count) + (pno*sizeof(DiskPage)),ios::beg);
	file.write((char*) &d, sizeof(DiskPage));
	check(file.rdstate() == ios::goodbit,"df::write");
};

int	df::alloc(int info = 0)
{
	DFPage	page;

	page.free = 0;
	page.info = info;
	write(page_count,page);
	return page_count++;
}

Buff*	df::load(int pno)
{
	int	i;
	Buff	*b1, *b2;

	for (i = 0, b1 = b2 = pool; i < DFPoolSize; ++i, ++b1)
		if (b1->pno == pno) {
			b1->access = access++;
			return b1;
		}
		else b2 = (b1->access < b2->access) ? b1 : b2;
	if (b2->modified)
		write(b2->pno,b2->page);
	b2->access = access++;
	b2->modified = False;
	b2->pno = pno;
	read(pno,b2->page);
	return b2;
}

//	Match	determines whether a record matches a saf.  The definition is deliberately verbose,
//	-----	in an attempt to minimise the number of recursive calls needed for its evaluation.

int	df::match(register field* rec, register field* saf)
{
	switch (saf->tag()) {
	case Size1 :
	case Size2 :
		if (rec->tag() != saf->tag())
			return False;
		register field *end1 = rec + rec->size();
		register field *end2 = saf + saf->size();
		for (++rec, ++saf; rec < end1 && saf < end2; rec += rec->size(), saf += saf->size())
			switch (saf->tag()) {
			case Size1 :
			case Size2 :
			 	if (!match(rec,saf))
					return False;
				break;
			default :
				if ((saf->tag() != Any) && (*rec != *saf))
					return False;
				break;
			}
		return rec == end1 && saf == end2;
	default :
		return (saf->tag() == Any) || (*rec == *saf);
	}
}

int	df::copy(field* rec1, field* rec2)
{
	register int	i = rec2->size();
		 int	s = rec2->size();

	while (i--) *rec1++ = *rec2++;
	return s;
}

int	df::insert(int pno, field *rec)
{
	Buff	*b = load(pno);
	int	free = b->page.free;

//	check(rec->size() < DFDataSize,"df::insert");
	if ((free + rec->size()) < DFDataSize) {
		copy(&b->page.data[free],rec);
		b->page.free += rec->size();
		b->modified = True;
		return True;
	}
	else	return False;
}

int	df::remove(int pno, field *saf)
{
	Buff	*b = load(pno);
	int	free = b->page.free;
	int	i, j, sz;

	for (i = j = 0; i < free; i += sz) {
		sz = b->page.data[i].size();
		if (!match(&b->page.data[i],saf))
			j += copy(&b->page.data[j],&b->page.data[i]);
	}
	if (j < free) {
		b->modified = True;
		b->page.free = j;
		return True;
	}
	else	return False;
}

field*	df::fetch(int pno, field *saf, field *res, field *end)
{
	Buff	*b = load(pno);
	int	free = b->page.free;

	for (int i = 0; i < free; i += b->page.data[i].size())
		if (match(&b->page.data[i],saf))
			if ((res + b->page.data[i].size()) >= end) {
				cerr << "df::fetch - virtual file too small" << endl;
				return end;
			}
			else {
				copy(res,&b->page.data[i]);
				res += res->size();
			}
	return res;
}

