# pragma implementation

# include	"../../header.h"
# include	"../tc/tc.h"
# include	"../df/df.h"
# include	"hf.h"
# include	"../db.h"

int	key_value(field f)
{
	extern db *database;

	return (f.tag() != Con) || (strncmp("Tuple",database->id_to_str(f.value()),strlen("Tuple")));
}

	hf::hf(df& data, const char* dname, const char* fname)
{
	char	string[MaxString];

	sprintf(string,"%s/curr/%s",dname,fname);
	file.open(string,(ios::in)|(ios::out)|(ios::nocreate));
	if (file.rdstate() == ios::goodbit) {
		file.read((char*) &global, sizeof(global));
		file.read((char*) &entries, sizeof(entries));
		file.read((char*) (directory = new int[entries]), sizeof(int)*entries);
	}
	else {
		file.open(string,(ios::in)|(ios::out));
		global =	0;
		entries = 	1;
		directory =	new int[1];
		directory[0] =	data.alloc(0);
	}
}

	hf::~hf()
{
	file.seekp(0,ios::beg);
	file.write((char*) &global, sizeof(global));
	file.write((char*) &entries, sizeof(entries));
	file.write((char*) directory, sizeof(int)*entries);
	delete[] directory;
	file.close();
};

int	hf::split()
{
	int *temp = new int[entries*2];

	if (!temp)
		return False;
	for (int i = 0; i < entries; ++i)
		temp[i+entries] = temp[i] = directory[i];
	delete[] directory;
	directory = temp;
	++global;
	entries *= 2;
	return True;
}

void	hf::page_split(df& data, int p1, int p2, int pivot)
{
	field	*end, *ptr, saf[1];

	for (register int i = 0; i < entries; ++i)
		if ((directory[i] == p1) && (i & pivot))
			directory[i] = p2;
	saf[0].set(Any,0);
	end = data.fetch(p1,saf,workspace,&workspace[DFDataSize]);
	for (ptr = workspace; ptr < end; ptr += ptr->size())
		if (hash_record(ptr,False) & pivot) {
			data.remove(p1,ptr);
			data.insert(p2,ptr);
		}
}

unsigned	hf::hash_record(field* rec, int pattern)
{
	register int	i, ct;
	field*		end;
	unsigned	code[MaxArgs], res;

	switch (rec->tag()) {
	case Size1 :
	case Size2 :
		end =  rec + rec->size();
		for (ct = 0, ++rec; rec < end; rec += rec->size())
			if (key_value(*rec))
				code[ct++] = hash_record(rec,pattern);
		for (i = res = 0; i < 32; ++i) {
			res |= (code[i%ct] & 0x1) << i;
			code[i%ct] >>= 1;
		}
		return res;
	case Any :
		return 0x00000000;
	default :
		return pattern ? 0xFFFFFFFF : rec->value();
	}
}

int	hf::insert(df& data, field* rec)
{
	int	key = hash_record(rec,False), pno = directory[key & mask()], local;

	if ((data.fetch(pno,rec,workspace,&workspace[DFDataSize])) != workspace)
		return False;
	for ( ; !data.insert(pno,rec); pno = directory[key & mask()]) {
		if (((local = data.ginfo(pno)) == global) && !split())
			return False;
		data.pinfo(pno,local+1);
		page_split(data,pno,data.alloc(local+1),0x1 << local);
	}
	return True;
}

field*	hf::fetch(df& data, field* saf, field* results, field* end)
{
	unsigned	rel =  hash_record(saf,True);
	unsigned	spec = rel & mask() & hash_record(saf,False);

	for (unsigned i = 0; i < entries; ++i)
		if ((i & rel) == spec)
			pages.push(directory[i]);
	for (field *ptr = results; !pages.empty(); )
		if ((ptr = data.fetch(pages.pop(),saf,ptr,end)) >= end)
			return end;
	return ptr;
}

int	hf::present(df& data, field* saf)
{
	unsigned rel =  hash_record(saf,True);
	unsigned spec = rel & mask() & hash_record(saf,False);
	for (unsigned i = 0; i < entries; ++i)
		if ((i & rel) == spec)
			pages.push(directory[i]);
	for (field *ptr = workspace; !pages.empty(); )
		if ((ptr = data.fetch(pages.pop(),saf,ptr,&workspace[DFDataSize])) != workspace) {
			for (i = 0; i < workspace[0].size(); ++i)
				saf[i] = workspace[i];
			return True;
		}
	return False;
}

int	hf::remove(df& data, field* saf)
{
	unsigned rel =  hash_record(saf,True);
	unsigned spec = rel & mask() & hash_record(saf,False);
	for (unsigned i = 0; i < entries; ++i)
		if ((i & rel) == spec)
			pages.push(directory[i]);
	for (int res = False; !pages.empty(); res |= data.remove(pages.pop(),saf));
	return res;
}

