# pragma implementation

# include	"../header.h"
# include	"../db/tc/tc.h"
# include	"../db/df/df.h"
# include	"../db/hf/hf.h"
# include	"../db/db.h"
# include	"../hp/cell.h"
# include	"../hp/cstack.h"
# include	"../hp/hp.h"
# include	"../user/user.h"
# include	"../expr/expr.h"
# include	"../type/type.h"

# include	"insert.C"
# include	"remove.C"


//	pending		Used for retrieval from the Pending file.
//	-------

int	expr::pending(char *file, cell* arg)
{
	field	saf[MaxCells];

	(void) ast_rec(arg,saf);
	for (int i = 0; i < MaxCells; ++i)
		if (saf[i] == AnyConst)
			saf[i].set(Any,0);
	if (database->present(file,saf)) {
		database->remove(file,saf);
		rec_ast(saf);
		return True;
	}
	else	return False;
}


//	fetch_sel	Pushes all tuples in file which match arg onto the heap stack.  Used for
//	---------	the implementation of selectors.

int	expr::fetch_sel(char *file, cell* arg)
{
	int	i, ct;
	field	saf[MaxCells], *start, *end;

	(void) ast_rec(arg,saf);
	for (i = 0; i < MaxCells; ++i)
		if (saf[i] == AnyConst)
			saf[i].set(Any,0);
	end = database->fetch(file,saf,records,records+heap_size);
	if (((end - records) < heap_size) && heap->allocate(end - records)) {
		for (start = records, ct = 0; start < end; start += start->size(), ++ct)
			rec_ast(start);
		return ct;
	}
	else	return -1;
}


//	fetch_exps	Pushes all the expressions associated with val in file onto the heap
//	----------	stack.  Used to retrieve the equations and types of functions.

int	expr::fetch_exps(char* file, field val)
{
	int	ct;
	field	saf[3], *start, *end;

	saf[0].set(Size1,3);
	saf[1] = val;
	saf[2].set(Any,0);
	end = database->fetch(file,saf,records,records+heap_size);
	if ((end - records) < heap_size) {
		for (start = records, ct = 0; start < end; start += start->size(), ++ct)
			rec_ast(start+2);
		return ct;
	}
	else	exit(2);
}

void	expr::set_dependency(field fun, field *rec, int incr)
{
	field	saf[4];

	for (int i = 0; i < rec->size(); ++i)
		if (rec[i].tag() == Con || rec[i].tag() == Var || rec[i].tag() == Sel1 || rec[i].tag() == Sel2) {
			saf[0].set(Size1,4);
			saf[1] = fun;
			saf[2] = rec[i];
			saf[3].set(Any,0);
			if (database->present("Depends",saf))
				database->remove("Depends",saf);
			saf[3].set(Int,saf[3].value() + incr);
			if (saf[3].value())
				database->insert("Depends",saf);
		}
}

field*	expr::ast_rec(cell* root, field* record)
{
	field	*t;

	if ((root->tag() == App) || (root->tag() == Lam)) {
		t = &record[1];
		t = ast_rec(root->left(),t);
		t = ast_rec(root->right(),t);
		record[0].set(root->tag() == App ? Size1 : Size2,t-record);
		return t;
	}
	record[0] = root->data();
	return ++record;
}

void	expr::rec_ast(field* rec)
{
	if (rec->tag() == Size1 || rec->tag() == Size2) {
		rec_ast(rec+1);				//	lhs of an abstraction or application
		rec_ast((rec+1) + ((rec+1)->size()));	//	rhs of an abstraction or application
		cell* rhs = heap->release();
		cell* lhs = heap->release();
		if (rec->tag() == Size1)
			heap->claim(heap->make(App,0,lhs,rhs));
		else	heap->claim(heap->make(Lam,0,lhs,rhs));
	}
	else	heap->claim(heap->make(*rec));
}

field*	expr::compile_ast(cell* ptr, field* inst)
{
	field	*curr = inst;

	for (int i = 0; ptr->tag() == Lam; ++i, ptr = ptr->right())
		curr = compile_arg(ptr->left(),curr);
	curr->set(Load,i);
	curr = compile_rhs(ptr,inst,curr,curr+1);
	curr->set(Load,0);
	return ++curr;
}

field*	expr::compile_arg(cell* ptr, field* inst)
{
	if (ptr->tag() == App) {
		inst = compile_arg(ptr->left(),inst);
		inst = compile_arg(ptr->right(),inst);
	}
	else {
		*inst = ptr->data();
		++inst;
	}
	return inst;
}

field*	expr::compile_rhs(cell* ptr, field* args, field* last, field* curr)
{
	int	i, j;

	if (ptr->tag() == App || ptr->tag() == Lam) {
		curr = compile_rhs(ptr->left(),args,last,curr);
		curr = compile_rhs(ptr->right(),args,last,curr);
		*curr = ptr->data();
		return ++curr;
	}
	else if (ptr->tag() == Var) {
		for (i = 0, j = -1; args < last; i += args->tag() == Var, ++args)
			if (ptr->data() == *args)
				j = i;
		if (j != -1)
			curr->set(Arg,j);	//	the variable is bound
		else	*curr = ptr->data();	//	the variable is free
		return ++curr;
	}
	else {
		*curr = ptr->data();
		return ++curr;
	}
}

int	expr::load(char* file, field fun, int sequence)
{
	cell	*array[MaxDefinitions], *temp;
	register int	i, j, top;

	if (!((i = fetch_exps(file,fun)) < MaxDefinitions)) {
		cout << "<< too many equations for `" << heap->make(fun) << "' >>" << endl;
		return next;
	}
	for (top = 0; i; --i, ++top)
		array[top] = heap->release();
	if (sequence)
		for (i = 0; i < top; ++i)
			for (j = i + 1; j < top; ++j) {
				switch (order(array[i],array[j])) {
				case Less :
					temp = array[i];
					array[i] = array[j];
					array[j] = temp;
					break;
				case Same :
					cout << "<< the equations for `" << heap->make(fun) << "' are ambiguous >>" << endl;
					return next;
				case More :
				case Incomparable :
					break;
				}
			}
	if ((i+next) >= MaxFunctions) {
		cout << "<< insufficient space to load equations - increase MaxFunctions >>" << endl;
		return next;
	}
	for (i = 0; i < top; ++i) {
		fun_name[i+next] = fun;
		fun_locn[i+next] = ptr;
		ptr = compile_ast(array[i],ptr);
	}
	next += i;
	return next - i;
}

int	expr::order(cell* e1,cell* e2)
{
	for (int result = Same; (e1->tag() == Lam) && (e2->tag() == Lam) ; e1 = e1->right(), e2 = e2->right())
		if ((result = order_args(e1->left(),e2->left())) != Same)
			return result;
	return result;
}

int	expr::order_args(cell* e1,cell* e2)
{
	cstack	spine1(MaxArgs), spine2(MaxArgs);
	int	result = Same;

	if ((e1->tag() == Var) && (e2->tag() == Var))
		return Same;
	if ((e1->tag() == Var) && (e2->tag() != Var))
		return Less;
	if ((e1->tag() != Var) && (e2->tag() == Var))
		return More;
	spine1.unwind(e1);		//	Both e1 and e2 are constructed values
	spine2.unwind(e2);
	if (spine1.pop()->data() != spine2.pop()->data())
		return Incomparable;
	while (spine1.used() && spine2.used())
		if ((result = order_args(spine1.pop()->right(),spine2.pop()->right())) != Same)
			return result;
	return result;
}

field	*expr::dependencies(field fun, field results[], field *end, int down, field* next)
{
	field	saf[4], recs[4*MaxDepth];
	field	*p1, *p2, *p3, *top, dep;

	if (!next)
		next = results;
	for (next = results, *next = fun, top = next+1; (next < top) && (top < end); ++next) {
		saf[0].set(Size1,4);
		saf[3].set(Any,0);			//	count of usage
		saf[1] = down ? *next : saf[3];		//	incorporate up info
		saf[2] = down ? saf[3] : *next;
		p2 = database->fetch("Depends",saf,recs,&recs[4*MaxDepth]);
		for (p1 = recs; (p1 < p2) && (top < end); p1 += 4) {
			dep = down ? p1[2] : p1[1];
			for (p3 = results; p3 < top && *p3 != dep; ++p3);
			if (p3 == top && !dep.simple())
				*top++ = dep;
		}
	}
	return top;
}
