//	include(s,t)		adds tuple t to selector s.
//	------------

int	expr::include(field sel, cell* tuple)
{
	char	name[MaxString];

	heap->claim(tuple);
	(void) ast_rec(tuple,records);
	heap->release();
	set_dependency(sel,records,1);
	strcpy(name,database->id_to_str(sel.value()));
	return database->insert(name,records);
}


//	selector(s,t)		The type of every function (recursively) dependent upon s is 
//	-------------		set to be unconstrained; the type of s is then asserted to be t

int	expr::selector(field sel, cell* t)
{
	remove_dependent(sel);
	return declare_type(sel,t);
}

//	function(f,t)		The type of every function (recursively) dependent upon f is 
//	-------------		set to be unconstrained; the type of f is then asserted to be t

int	expr::function(field fun, cell* t)
{
	remove_dependent(fun);
	return declare_type(fun,t);
}


//	declare_type(f,t)	asserts that the new type of f is t.
//	-----------------

int	expr::declare_type(field f, cell* t)
{
	records[0].set(Size1,3);			//	remove the old type
	records[1] = f;
	records[2].set(Any,0);
	database->present("Types",records);
	set_dependency(f,&records[2],-1);
	database->remove("Types",records);
	database->insert("Pending",&f);
	if (t != ((cell*) 0)) {				//	insert the new type
		(void) ast_rec(t,&records[2]);
		records[0].set(Size1,records[2].size()+2);
		records[1] = f;
		set_dependency(f,&records[2],1);
		return database->insert("Types",records);
	}
	else	return True;
}


//	insert_eqn(f,e)		adds an equation e for function f.
//	---------------		Currently e must be protected before the function is called :
//				this must be fixed !

int	expr::insert_eqn(field fun, cell* eqn)
{
	cstack	args(MaxArgs);
	cell	*bdy, *arg;

	if (!heap->allocate(MaxArgs))
		return False;
	eqn = replace_abst(eqn);
	for (bdy = eqn; bdy->tag() == Lam; bdy = bdy->right())
		args.push(bdy->left());
	if (!heap->allocate(MaxCells))
		return False;
	for (bdy = replace_zf(bdy,args); arg = args.pop(); bdy = heap->make(Lam,0,arg,bdy));
// cout << "Function " << heap->make(fun) << " = " << bdy << endl;
	(void) ast_rec(bdy,&records[2]);
	records[0].set(Size1,records[2].size()+2);
	records[1] = fun;
	if (!database->insert("Equations",records))
		return False;
	database->insert("Pending",&fun);
	set_dependency(fun,&records[2],1);
	return True;
}

cell*	expr::replace_abst(cell* eqn)
{
	cstack	spine(MaxArgs);
	cell	*tip = spine.unwind(eqn), *lhs, *rhs;

	if (tip->data() == Define)
		return eqn;
	if (tip->data() == LambdaConst) {
		spine.pop();
		lhs = spine.pop()->right();
		rhs = spine.pop()->right();
		return heap->make(Lam,0,lhs,replace_abst(rhs)); 
	}
	for (lhs = spine.pop(); rhs = spine.pop(); lhs = heap->make(App,0,lhs,replace_abst(rhs->right())));
	return lhs;
}

cell*	expr::replace_zf(cell* eqn, const cstack& argin)
{
	field	sub_fun, saf[1];
	char	string[MaxString];
	cell	*patt, *def;
	cstack  args = argin;

	switch (eqn->tag()) {
	case App :
		return heap->make(App,0,replace_zf(eqn->left(),args),replace_zf(eqn->right(),args));
	case Lam :
		saf[0].set(Any,0);
		database->present("NewFunctions",saf);
		database->remove("NewFunctions",saf);
		saf[0].set(Int,saf[0].value() + 1);
		database->insert("NewFunctions",saf);
		sprintf(string,"_%d",saf[0].value());
		sub_fun.set(Var,database->str_to_id(string));
		args.push(eqn->left());
		eqn = replace_zf(eqn->right(),args);
		patt = args.pop();
		eqn = heap->make(Lam,0,patt,eqn);
		if (patt->tag() != Var) {		//	Add a default equation
			def = heap->make(Lam,0,heap->make(Var,database->str_to_id("-var-")),heap->make(NilConst));
			make_application(sub_fun,def,args,eqn);
		}
		return	make_application(sub_fun,eqn,args,eqn);
	default :
		return eqn;
	}
}

cell*	expr::make_application(field sub_fun, cell* eqn, const cstack& argin, cell* bdy)
{
	cell	*arg, *app;
	cstack	args1(MaxArgs), args2(MaxArgs);
	cstack  args = argin;

	for (args1 = args; arg = args1.pop(); ) {
		switch (arg->tag()) {
		case App :
			for (; arg->tag() == App; arg = arg->left())
				args1.push(arg->right());
			break;
		case Var :
			if (occurs_in(arg,bdy)) {
				for (int i = 0; i < args2.used() && !occurs_in(arg,args2[i]); ++i);
				if (i == args2.used()) {
					args2.push(arg);
					eqn = heap->make(Lam,0,arg,eqn);
				}
			}
			break;
		}
	}
	for (app = heap->make(sub_fun); arg = args2.pop(); app = heap->make(App,0,app,arg));
	insert_eqn(sub_fun,eqn);
	return app;
}

int	expr::occurs_in(cell* val, cell* exp)
{
	switch (exp->tag()) {
	case App :
		return occurs_in(val,exp->left()) || occurs_in(val,exp->right());
	case Lam :
		return (!occurs_in(val,exp->left())) && (occurs_in(val,exp->right()));
	default :
		return exp->data() == val->data();
	}
}
