# pragma implementation

# include	"../header.h"
# include	"../db/tc/tc.h"
# include	"../db/df/df.h"
# include	"../db/hf/hf.h"
# include	"../db/db.h"
# include	"../hp/cell.h"
# include	"../hp/cstack.h"
# include	"../hp/hp.h"
# include	"../expr/expr.h"
# include	"../type/type.h"
# include	"../user/user.h"
# include	"../eval/eval.h"
# include	"oper.h"

field	CloseR;
field	CloseW;
field	Commit;
field	Define;
field	Done;
field	Exclude;
field	Get;
field	Include;
field	OpenR;
field	OpenW;
field	Peek;
field	Print;
field	Put;
field	Restore;

field	TypeConst;
field	TrueConst;
field	FalseConst;
field	ConsConst;
field	NilConst;

field	WrongConst;
field	FunConst;
field	ListConst;
field	OperConst;
field	AnyConst;
field	LambdaConst;

field	FlatmapVar;
field	IfVar;

field	Query;

	oper::oper(char* nm)
{
	strcpy(name,nm);
	database = new db(name);
	in_open = out_open = False;

	CloseR.set(Con,database->str_to_id("CloseR"));
	CloseW.set(Con,database->str_to_id("CloseW"));
	Commit.set(Con,database->str_to_id("Commit"));
	Define.set(Con,database->str_to_id("Define"));
	Done.set(Con,database->str_to_id("Done"));
	Exclude.set(Con,database->str_to_id("Exclude"));
	Get.set(Con,database->str_to_id("Get"));
	Include.set(Con,database->str_to_id("Include"));
	OpenR.set(Con,database->str_to_id("OpenR"));
	OpenW.set(Con,database->str_to_id("OpenW"));
	Peek.set(Con,database->str_to_id("Peek"));
	Print.set(Con,database->str_to_id("Print"));
	Put.set(Con,database->str_to_id("Put"));
	Restore.set(Con,database->str_to_id("Restore"));

	TypeConst.set(Con,database->str_to_id("Type"));
	WrongConst.set(Con,database->str_to_id("Wrong"));
	FunConst.set(Con,database->str_to_id("(->)"));
	ListConst.set(Con,database->str_to_id("List"));
	OperConst.set(Con,database->str_to_id("Oper"));
	AnyConst.set(Con,database->str_to_id("Any"));
	LambdaConst.set(Con,database->str_to_id("/\\"));

	TrueConst.set(Con,database->str_to_id("True"));
	FalseConst.set(Con,database->str_to_id("False"));
	ConsConst.set(Con,database->str_to_id("(:)"));
	NilConst.set(Con,database->str_to_id("[]"));

	FlatmapVar.set(Var,database->str_to_id("flatmap"));
	IfVar.set(Var,database->str_to_id("if"));

	Query.set(Var,database->str_to_id("_query"));
}


//	process(c)	Causes the command c to be executed.
//	----------

void	oper::process(cell* qry)
{
	cstack	spine(MaxArgs);
	cell	*res;
	field	oper, command;

	command.set(Var,database->str_to_id("_command"));
	expressions->insert_eqn(command,qry);
	qry = heap->make(command);
	typecheck = True;
	while (!interrupt && confirm_oper(qry) && evaluate(qry) && qry->data() != Done) {
		expressions->unload();
		heap->claim(qry);			//	protect the query
		spine.unwind(qry);
		oper = qry->tip()->data();
		if (oper == Define) {
			res = heap->make(Int,define(spine[2]->right(),spine[1]->right()));
			typecheck = True;
		}
		else if (oper == Exclude) {
			res = heap->make(Int,update(spine[2]->right()->data(),spine[1]->right(),False));
		}
		else if (oper == Include) {
			res = heap->make(Int,update(spine[2]->right()->data(),spine[1]->right(),True));
		}
		else if (oper == Print) {
			res = heap->make(Int,show(spine[1]->right()));
		}
		else if (oper == Commit) {
			res = heap->make(Int,filemanager("Commit"));
		}
		else if (oper == Restore) {
			res = heap->make(Int,filemanager("Restore"));
			typecheck = True;
		}
		else if (oper == CloseR) {
			res = heap->make(Int,closer());
		}
                else if (oper == CloseW) {
			res = heap->make(Int,closew());
                }
                else if (oper == OpenR) {
			res = heap->make(Int,openr(spine[1]->right()));
                }
		else if (oper == OpenW) {
			res = heap->make(Int,openw(spine[1]->right()));
		}
		else if (oper == Get) {
			char	c;
			res = heap->make(Int,get(c));
			res = heap->make(App,0,heap->make(App,0,heap->make(Con,database->str_to_id("Tuple2")),res),heap->make(Chr,c));
		}
		else if (oper == Peek) {
			char	c;
			res = heap->make(Int,peek(c));
			res = heap->make(App,0,heap->make(App,0,heap->make(Con,database->str_to_id("Tuple2")),res),heap->make(Chr,c));
		}
		else if (oper == Put) {
			res = heap->make(Int,put(spine[1]->right()));
		}
		heap->claim(res);
		if (!heap->allocate(4))
			interrupt = True;
		else {
			qry = heap->make(App,0,spine[0]->right(),res);
			heap->release();			//	unprotect the result
			heap->release();			//	unprotect the query
		}
	}
	while (heap->release());
	expressions->remove_eqns(command);
	return;
}

int	oper::confirm_oper(cell* op)
{
	cell	*t;

	if (typecheck) {
		typecheck = False;
		heap->claim(op);		//	claim operation
		types.cascade();
		t = types.fetch(op,True);
		if (t->data() != OperConst) {
			cerr << "<< the command is of type `" << t << "' rather than of type `Oper' >>" << endl << flush;
			heap->release();
			return False;
		}
		heap->release();		//	release operation
	}
	return True;
}

int	oper::confirm_patt(cell* exp)
{
	switch (exp->tag()) {
	case Chr :
	case Con :
	case Str :
	case Int :
	case Real :
	case Var :
		return True;
	default :
		for (; exp->tag() == App; exp = exp->left())
			if (!confirm_patt(exp->right()))
				return False;
		return exp->tag() == Con;
	}
}

int	oper::confirm_unused(field f)
{
	field	saf[4];

	saf[0].set(Size1,4);
	saf[1].set(Any,0);
	saf[2] = f;
	saf[3].set(Any,0);
	if (!database->present("Depends",saf))
		return True;
	cout << "<< `" << database->id_to_str(saf[2].value()) << "' is referenced by `" << database->id_to_str(saf[1].value()) << "' (and possibly elsewhere) >> " << endl << flush;
	return False;
}

# include	"closer.C"
# include	"closew.C"
# include	"define.C"
# include	"display.C"
# include	"extsel.C"
# include	"filemanager.C"
# include	"function.C"
# include	"get.C"
# include	"intsel.C"
# include	"openr.C"
# include	"openw.C"
# include	"peek.C"
# include	"put.C"
# include	"type.C"
# include	"update.C"
# include	"value.C"
