#pragma implementation
# include	"../header.h"
# include	"../db/tc/tc.h"
# include	"../db/df/df.h"
# include	"../db/hf/hf.h"
# include	"../db/db.h"
# include	"../hp/cell.h"
# include	"../hp/cstack.h"
# include	"../hp/hp.h"
# include	"../user/user.h"
# include	"../expr/expr.h"
# include	"type.h"

const	int	MaxType =	100;

int	type::existential(cell* c)
{
	while (c->tip()->data() == FunConst)
		c = c->right();
	return c->tag() == Con;
}

int	type::universal(cell* c)
{
	cstack	pending(MaxArgs);
	cell	*t;

	for (t = c; t->tip()->data() == FunConst; t = t->right())
		pending.push(t->left()->right());
	while (c = pending.pop()) {
		for (; c->tag() == App; c = c->left())
			pending.push(c->right());
		if (c->tag() == TVar)
			if (!expressions->occurs_in(c,t))
				return False;
	}
	return True;
}

int	type::first_order(cell* t)
{
	return (t->tag() == App) ? (first_order(t->left()) && first_order(t->right())) : (t->data() != FunConst);
}

int	type::monomorphic(cell* t)
{
	return (t->tag() == App) ? (monomorphic(t->left()) && monomorphic(t->right())) : (t->tag() == Con);
}

int	type::size(cell* exp)
{
	exp = value(exp);
	return (exp->tag() == App) ? (1 + size(exp->left()) + size(exp->right())) : 1;
}

int	type::occurs(field var, cell* exp)
{
	exp = value(exp);
	return (exp->tag() == App) ? (occurs(var,exp->left()) || occurs(var,exp->right())) : (var == exp->data());
}

cell*	type::standardise(cell* exp, int exist, cstack& s1, cstack& s2)
{
	switch (exp->tag()) {
	case App :
		return heap->make(App,0,standardise(exp->left(),exist,s1,s2),standardise(exp->right(),exist,s1,s2));
	case TVar :
		for (int i = 0; i < s1.used(); ++i)
			if (exp->data() == s1[i]->data())
				return s2[i];
		s1.push(exp);
		s2.push(exist ? heap->make(TCon,new_constructor()) : heap->make(TVar,new_variable()));
		return s2[i];
	default :
		return exp;
	}
}

int	type::unify(cell* e1, cell* e2)
{
	if ((e1 = value(e1))->data() == (e2 = value(e2))->data())
		return (e1->tag() != App) || (unify(e1->left(),e2->left()) && unify(e1->right(),e2->right()));
	if (e1->tag() == TVar && !occurs(e1->data(),e2))
		return substitute(e1->data(),e2);
	if (e2->tag() == TVar && !occurs(e2->data(),e1))
		return substitute(e2->data(),e1);
	return False;
}

cell*	type::fetch(cell* e, int reset)
{
	int	base = heap->used();

	if (reset) {
		nextv = 0;
		reset_env((bind*) 0);
	}
	if (infer(e,False)) {
		fix(e = heap->release());
		return e;
	}
	for (;base > heap->used(); heap->release());
	return heap->make(WrongConst);
}

int	type::infer(cell* e, int binding)
{
	cell	*t1, *t2;
	char	name[MaxString];

	if (!heap->allocate(MaxCells)) {
		cout << "<< insufficient heap space to infer a type for `" << e << "' >>" << endl << flush;
		return False;
	}
	switch (e->tag()) {
	case TVar :
		heap->claim(heap->make(Con,database->str_to_id("Type")));
		return True;
	case Int :
	case Real :
		heap->claim(heap->make(Con,database->str_to_id("Num")));
		return True;
	case Chr :
		heap->claim(heap->make(Con,database->str_to_id("Chr")));
		return True;
	case Str :
		heap->claim(heap->make(Con,database->str_to_id("Str")));
		return True;
	case Any :
		heap->claim(heap->make(TVar,new_variable()));
		return True;
	case Var :
		if (binding) {
			substitute(e->data(),t1 = heap->make(TVar,new_variable()));
			heap->claim(t1);
			return True;
		}
		for (bind* ptr = env; ptr; ptr = ptr->next)
			if (ptr->var == e->data()) {
				heap->claim(ptr->exp);
				return True;
			}
		if (expressions->fetch_exps("Types",e->data()))
			heap->claim(standardise(heap->release()));
		else	heap->claim(heap->make(TVar,new_variable()));
		return True;
	case Con :
		if (!expressions->fetch_exps("Types",e->data())) {
			cout << "<< undeclared value `" << e << "' >>" << endl << flush;
			return False;
		}
		t1 = heap->release();
		heap->claim(standardise(t1,binding && existential(t1)));
		return True;
	case Pre :
	case Sel1 :
		if (expressions->fetch_exps("Types",e->data())) {
			heap->claim(standardise(heap->release()));
			return True;
		}
		cout << "<< undeclared value `" << e << "' >>" << endl << flush;
		return False;
	case Sel2 :
		strcpy(name,database->id_to_str(e->value()));
		heap->claim(heap->make(App,0,heap->make(ListConst),heap->make(Con,database->str_to_id(&name[3]))));
		return True;
	case Lam :
		if (!infer(e->left(),True) || !infer(e->right(),False))
			return False;
		heap->allocate(3);
		t2 = heap->release();
		t1 = heap->release();
		heap->claim(heap->make(App,0,heap->make(App,0,heap->make(FunConst),t1),t2));
		return True;
	case App :
		if (!infer(e->left(),binding) || !infer(e->right(),binding))
			return False;
		heap->allocate(4);
		t2 = heap->release();
		t1 = heap->release();
		t2 = heap->make(App,0,heap->make(App,0,heap->make(FunConst),t2),heap->make(TVar,new_variable()));
		fix(t1);
		fix(t2);
		if (unify(t1,t2)) {
			fix(t2);
			heap->claim(t2->right());
			return True;
		}
		cout << "<< type error --> " << e << " <--" << endl;
		cout << "\t`" << e->left() << "' has type `" << t1 << "'" << endl;
		cout << "\t`" << e->right() << "' has type `" << t2->left()->right() << "'" << endl;
		cout << ">>" << endl << flush;
		return False;
	}
	cout << "Illegal tag value encountered" << endl;
	return False;
}

int	type::fix(cell* old_type)
{
	cell	*new_type = value(old_type);
	int	change = new_type != old_type;

	*old_type = *new_type;
	if (old_type->tag() == App) {
		change |= fix(old_type->left());
		change |= fix(old_type->right());
	}
	return change;
}

void	type::cascade()
{
	cell	*exp;

	while (expressions->pending("Pending",heap->make(Any,0))) {
		heap->claim(exp = heap->release());	//	protect the expression
		switch (exp->tag()) {
		case Sel1 :
			cascade_up(exp);
			break;
		case Sel2 :
			cascade_intsel(exp);
			break;
		case Var :
			cascade_fun(exp);
			break;
		}
		heap->release();			//	release the expression
	}
}

void	type::cascade_intsel(cell* sel)
{
	cell	*reltype, *seltype, *rel;
	int	ct;

	heap->claim(seltype = fetch(sel,True));		//	infer a type for the selector
	ct = expressions->fetch_exps("Equations",sel->data());
	while (ct--) {
		heap->claim(rel = heap->release());	//	pick an equation for the selector
		if (seltype->data() != WrongConst) {
			reltype = fetch(rel,False);	//	infer a type for the equation
			if (!unify(reltype,seltype->right())) {
				cout << "<< type error --> " << sel << " <--" << endl << flush;
				cout << "\texisting type is `" << seltype << "'" << endl << flush;
				cout << "\trelation type is `" << reltype << "'" << endl << flush;
				cout << ">>" << endl << flush;
				seltype = heap->make(WrongConst);
				expressions->declare_type(sel->data(),seltype);
				cascade_up(sel);
			}
			else if (fix(seltype)) {
				if (size(seltype) > MaxType) {
					cout << "<< the inferred type of `" << sel << "' is too complex >>" << endl << flush;
					expressions->declare_type(sel->data(),heap->make(WrongConst));
				}
				else	expressions->declare_type(sel->data(),seltype);
				cascade_up(sel);
			}
		}
		heap->release();			//	release the equation
	}
	heap->release();				//	release the selector type
}

void	type::cascade_fun(cell* fun)
{
	cell	*eqntype, *funtype, *eqn;
	int	change = False, ct;

	heap->claim(funtype = fetch(fun,True));			//	infer a type for the function
	ct = expressions->fetch_exps("Equations",fun->data());
	while (ct-- > 0) {
		heap->claim(eqn = heap->release());		//	pick an equation for the function
		if (funtype->data() != WrongConst) {
			eqntype = fetch(eqn,False);		//	infer a type for the equation
			if (!unify(eqntype,funtype)) {
				cout << "<< type error --> " << fun << " <--" << endl << flush;
				cout << "\texisting type is `" << funtype << "'" << endl << flush;
				cout << "\tequation type is `" << eqntype << "'" << endl << flush;
				cout << ">>" << endl << flush;
				change = True;
				funtype = heap->make(WrongConst);
				expressions->declare_type(fun->data(),funtype);
			}
			else if (fix(funtype)) {
				change = True;
				if (size(funtype) > MaxType) {
					cout << "<< the inferred type of `" << fun << "' is too complex >>" << endl << flush;
					expressions->declare_type(fun->data(),heap->make(WrongConst));
				}
				else	expressions->declare_type(fun->data(),funtype);
			}
		}
		heap->release();				//	release the equation
	}
	heap->release();					//	release the function type
	if (change)
		cascade_up(fun);
}

void	type::cascade_up(cell* fun)
{
	field	depends[MaxFunctions], *start, *end;

	if (strncmp("_",database->id_to_str(fun->value()),1) != 0)
		cout << "// `" << fun << "' now has type `" << fetch(fun,True) << "'" << endl;
	end = expressions->dependencies(fun->data(),depends,&(depends[MaxFunctions]),False);
	for (start = depends+1; start < end; ++start)
		database->insert("Pending",start);
}

